// Copyright (c) 2001, 2002 Zope Foundation and Contributors.
// All Rights Reserved.
//
// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE

package main
// Bits that should be in ZODB		XXX -> zodb

import (
	"context"
	"fmt"
	"sync"
	"sync/atomic"

	"lab.nexedi.com/kirr/neo/go/zodb"

	pickle "github.com/kisielk/og-rek"
)

// XXX make methods private, e.g. _pJar ?

// Object is the interface that every in-RAM object representing any ZODB object implements.
type Object interface {
	PJar()    *Connection	// Connection this in-RAM object is part of.
	POid()    zodb.Oid	// object ID in the database.

	// object serial as of database state for particular Connection (PJar).
	// 0 if not yet loaded (XXX ok?)
	PSerial() zodb.Tid


	// PActivate brings object to live state.
	//
	// It requests to persistency layer that in-RAM object data to be present.
	// If object state was not in RAM - it is loaded from the database.
	//
	// On successful return the object data is either the same as in the
	// database or, if this data was previously modified by user of
	// object's jar, that modified data.
	//
	// Object data must be accessed only after corresponding PActivate
	// call, which marks that object's data as being in use.
	PActivate(ctx context.Context) error

	// PDeactivate indicates that corresponding PActivate caller finished access to object's data.
	//
	// As PActivate makes sure object's data is present in-RAM, PDeactivate
	// tells persistency layer that this data is no longer used by
	// corresponding PActivate caller.
	//
	// Note that it is valid to have several concurrent uses of object
	// data, each protected with corresponding PActivate/PDeactivate pair:
	// as long as there is still any PActivate not yet compensated with
	// corresponding PDeactivate, object data will assuredly stay alive in RAM.
	//
	// Besides exotic cases, the caller thus must not use object's data
	// after PDeactivate call.
	PDeactivate()

	// PInvalidate requests in-RAM object data to be discarded.
	//
	// Irregardless of whether in-RAM object data is the same as in the
	// database, or it was modified, that in-RAM data must be forgotten.
	//
	// PInvalidate must not be called while there is any in-progress
	// object's data use (PActivate till PDeactivate).
	//
	// In practice this means that:
	//
	//	- application must make sure to finish all objects accesses
	//	  before transaction boundary: at transaction boundary - either
	//	  at abort or commit, the persistency layer will sync to
	//	  database and process invalidations.
	//
	//	- if PInvalidate is explicitly called by application, the
	//	  application must care to make sure it does not access the
	//	  object data simultaneously.
	PInvalidate()
}

// PyObject is the interface that every in-RAM object representing Python ZODB object implements.
type PyObject interface {
	Object

	PyClass() pickle.Class // python class of this object
//	PyState() interface{}  // object state. python passes this to pyclass.__new__().__setstate__()

	// XXX want to hide from PyObject. Rationale: we do not want e.g. PySetState to
	// be available to user who holds PyObject interface: it is confusing to have
	// both PActivate and PySetState at the same time.
//	PyStateful
}

// ObjectState describe state of in-RAM object.
type ObjectState int

const (
	GHOST ObjectState = -1
	UPTODATE          = 0
	CHANGED           = 1
	// no STICKY - we pin objects in RAM with PActivate
)

// object is common base for in-RAM implementations of ZODB objects.
type object struct {
	jar	*Connection
	oid	zodb.Oid
	serial	zodb.Tid

	mu	sync.Mutex
	state	ObjectState
	refcnt	int32
}

func (obj *object) PJar() *Connection	{ return obj.jar	}
func (obj *object) POid() zodb.Oid	{ return obj.oid	}
func (obj *object) PSerial() zodb.Tid	{ return obj.serial	}

// pyObject is common base for in-RAM implementations of ZODB Python objects.
type pyObject struct {
	object
	pyclass pickle.Class

	// protected by object.mu
	instance PyStateful
	loaded   *loaded
//	loaderr  error		// if there was error at state loading
//	ready    chan struct{}	// activation complete
}

func (pyobj *pyObject) PyClass() pickle.Class	{ return pyobj.pyclass	}
//func (pyobj *pyObject) PyState() interface{}	{ return pyobj.pystate	}

// loaded indicates object's load state/result.
//
// when !ready the loading is in progress.
// when ready the loading has been completed.
type loaded struct {
	ready chan struct{} // closed when loading finishes

	// error from the load.
	// if there was no error loaded data goes to object state.
	err   error
}

// XXX
type Stateful interface {
	// DropState should discard in-RAM object state.
	DropState()
}

// PyStateful is the interface describing in-RAM object whose data state can be
// exchanged as Python data.
type PyStateful interface {
	Stateful

	// PySetState should set state of the in-RAM object from Python data.
	// Analog of __setstate__() in Python.
	PySetState(pystate interface{}) error

	// PyGetState should return state of the in-RAM object as Python data.
	// Analog of __getstate__() in Python.
	//PyGetState() interface{}	XXX
}


// Connection represents a view of ZODB database.
//
// The view is representing state of ZODB objects as of `at` transaction.
//
// Connection changes are private and are isolated from changes in other Connections.
//
// XXX Connection, and {Py}Object methods that relate to it, are not safe for
// modifications from multiple goroutines simultaneously.
//
// XXX ^^^ better must be safe - use case: e.g. prefetch.
type Connection struct {
	stor	zodb.IStorage	// underlying storage
	at	zodb.Tid	// current view of database

	// {} oid -> obj
	//
	// rationale:
	//
	// on invalidations: we need to go oid -> obj and invalidate it.
	// -> Connection need to keep {} oid -> obj.
	// -> we can use that {} when loading a persistent Ref twice to get to the same object.
	//
	// however: if Connection keeps strong link to obj, just
	// obj.PDeactivate will not fully release obj if there are no
	// references to it from other objects:
	//
	//	- deactivate will release obj state (ok)
	//	- but there will be still reference from connection `oid -> obj` map to this object.
	//
	// -> we can solve it by using "weak" pointers in the map.
	//
	// NOTE we cannot use regular map and arbitrarily manually "gc" entries
	// there periodically: since for an obj we don't know whether other
	// objects are referencing it, we can't just remove obj's oid from
	// the map - if we do so and there are other live objects that
	// reference obj, user code can still reach obj via those
	// references. On the other hand, if another, not yet loaded, object
	// also references obj and gets loaded, traversing reference from
	// that loaded object will load second copy of obj, thus breaking 1
	// object in db <-> 1 live object invariant:
	//
	//	A  →  B  →  C
	//	↓           |
	//      D <--------- - - -> D2 (wrong)
	//
	// - A activate
	// - D activate
	// - B activate
	// - D gc, A still keeps link on D
	// - C activate -> it needs to get to D, but D was removed from objtab
	//   -> new D2 is wrongly created
	//
	// that's why we have to depend on Go's GC to know whether there are
	// still live references left or not. And that in turn means finalizers
	// and thus weak references.
	//
	// some link on the subject:
	// https://groups.google.com/forum/#!topic/golang-nuts/PYWxjT2v6ps
	//
	// NOTE2 finalizers don't run on when they are attached to an object in cycle.
	// Hopefully we don't have cycles with ZBtree/ZBucket	XXX verify this
	objmu  sync.Mutex
	objtab map[zodb.Oid]*WeakRef	// oid -> WeakRef(PyObject)


	// hooks for application to influence live caching decisions.
	cacheControl LiveCacheControl
}

// LiveCacheControl is the interface that allows applications to influence
// Connection's decisions with respect to its live cache.
type LiveCacheControl interface {
	// WantEvict is called when object is going to be evicted from live cache and made ghost.
	// If !ok the object will remain live.
	WantEvict(obj Object) (ok bool)
}


// ----------------------------------------


// Get returns in-RAM object corresponding to specified ZODB object according to current database view.
//
// If there is already in-RAM object that corresponds to oid, that in-RAM object is returned.
// Otherwise new in-RAM object is created and filled with object's class loaded from the database.
//
// The scope of the object returned is the Connection.	XXX ok?
//
// The object's data is not neccessarily loaded after Get returns. Use
// PActivate to make sure the object ifs fully loaded.
func (conn *Connection) Get(ctx context.Context, oid zodb.Oid) (interface{} /*PyObject*/, error) {
	conn.objmu.Lock()		// XXX -> rlock
	wobj := conn.objtab[oid]
	var xobj interface{}
	if wobj != nil {
		xobj = wobj.Get()
	}
	conn.objmu.Unlock()

	// object was already there in objtab.
	if xobj != nil {
		return xobj, nil
	}

	// object is not there in objtab - raw load it, get its class -> get(pyclass, oid)
	pyclass, pystate, serial, err := conn.loadpy(ctx, oid)
	if err != nil {
		return nil, err		// XXX errctx
	}

	xobj, err = conn.get(pyclass, oid)
	if err != nil {
		return nil, err
	}

	// XXX we are dropping just loaded pystate. Usually Get should be used
	// to only load root object, so maybe that is ok.
	//
	// TODO -> use (pystate, serial) to activate.
	_, _ = pystate, serial
	return xobj, nil
}


// wrongClassError is the error cause returned when object's class is not what was expected.
type wrongClassError struct {
	want, have pickle.Class
}

func (e *wrongClassError) Error() string {
	return fmt.Sprintf("wrong class: want %q; have %q", e.want, e.have)
}

// get returns in-RAM object corresponding to specified ZODB object according to current database view.
//
// If there is already in-RAM object that corresponds to oid, that in-RAM object is returned.
// Otherwise new in-RAM object is created according to specified class.
//
// The object's data is not neccessarily loaded after get returns. Use
// PActivate to make sure the object is fully loaded.
//
// XXX object scope.
//
// Use-case: in ZODB references are (pyclass, oid), so new ghost is created
// without further loading anything.
func (conn *Connection) get(pyclass pickle.Class, oid zodb.Oid) (PyObject, error) {
	conn.objmu.Lock()		// XXX -> rlock
	wobj := conn.objtab[oid]
	var pyobj PyObject
	checkClass := false
	if wobj != nil {
		if xobj := wobj.Get(); xobj != nil {
			pyobj = xobj.(PyObject)
		}
	}
	if pyobj == nil {
		pyobj = conn.newGhost(pyclass, oid)
		conn.objtab[oid] = NewWeakRef(pyobj)
	} else {
		checkClass = true
	}
	conn.objmu.Unlock()

	if checkClass {
		if cls := pyobj.PyClass(); pyclass != cls {
			return nil, &zodb.OpError{
				URL:  conn.stor.URL(),
				Op:   fmt.Sprintf("@%s: get", conn.at), // XXX abuse
				Args: oid,
				Err:  &wrongClassError{pyclass, cls},
			}
		}
	}

	return pyobj, nil
}


// loadpy loads object specified by oid and decodes it as a ZODB Python object.
//
// loadpy does not create any in-RAM object associated with Connection.
// It only returns decoded database data.
func (conn *Connection) loadpy(ctx context.Context, oid zodb.Oid) (pyclass pickle.Class, pystate interface{}, serial zodb.Tid, _ error) {
	buf, serial, err := conn.stor.Load(ctx, zodb.Xid{Oid: oid, At: conn.at})
	if err != nil {
		return pickle.Class{}, nil, 0, err
	}

	defer buf.Release()

	pyclass, pystate, err = zodb.PyData(buf.Data).Decode()
	if err != nil {
		return pickle.Class{}, nil, 0, err	// XXX err ctx
	}

	return pyclass, pystate, serial, nil
}

// ---- pyclass -> new ghost ----

// path(class) -> new(pyobj)
var classTab = make(map[string]func(*pyObject)PyObject)

// registerClass registers python class to be transformed to Go instance
// created via classNew.
//
// must be called from global init().
func registerClass(classPath string, classNew func(*pyObject)PyObject) {
	classTab[classPath] = classNew
}

// newGhost creates new ghost object corresponding to pyclass and oid.
func (conn *Connection) newGhost(pyclass pickle.Class, oid zodb.Oid) PyObject {
	pyobj := &pyObject{
		object:  object{jar: conn, oid: oid, serial: 0, state: GHOST},
		pyclass: pyclass,
	}

	// switch on pyclass and transform e.g. "zodb.BTree.Bucket" -> *ZBucket
	classNew := classTab[pyclass.Module + "." + pyclass.Name]
	var instance PyObject
	if classNew != nil {
		instance = classNew(pyobj)
	} else {
		instance = &dummyPyInstance{pyObject: pyobj}
	}

	pyobj.instance = instance
	return instance
}

// dummyPyInstance is used for python classes that were not registered.
type dummyPyInstance struct {
	*pyObject
	pystate interface{}
}

func (d *dummyPyInstance) DropState() {
	d.pystate = nil
}

func (d *dummyPyInstance) PySetState(pystate interface{}) error	{
	d.pystate = pystate
	return nil
}


// ----------------------------------------

// object's activate & friends that only manage base activation state, without actually loading data.

// activate increments object reference counter.
//
// it returns whether object data needs to be loaded.
// must be called with .mu held.
func (obj *object) activate() (load bool) {
	obj.refcnt++
	return (obj.refcnt == 1 && obj.state == GHOST)
/*
	nuse := atomic.AddInt32(&obj.refcnt, +1)
	if nuse == 1 {
		// we become responsible for loading object's data.
		// XXX also check state (it could be already loaded, but with refcnt=0)
		return true
	}
	return false
*/
}

// deactivate decrements object reference counter.
//
// it returns whether object data needs to be discarded.
// must be called with .mu held.
func (obj *object) deactivate() (drop bool) {
	//nuse := atomic.AddInt32(&obj.refcnt, -1)
	obj.refcnt--
	if obj.refcnt < 0 {
		panic("deactivate: refcnt < 0")
	}
	if obj.refcnt > 0 {
		return // users still left
	}

	// no users left. Let's see whether we should transition this object to ghost.
	drop = true
	if obj.state >= CHANGED {
		drop = false
	}

	// XXX -> pyObject?
	if drop {
		if cc := obj.jar.cacheControl; cc != nil {
			drop = cc.WantEvict(obj.instance)
		}
	}

	if drop {
		obj.serial  = 0
		obj.instance.DropState()
	}

	return drop
}


// PActivate implements Object.
func (pyobj *pyObject) PActivate(ctx context.Context) (err error) {
	pyobj.mu.Lock()
	doload := pyobj.activate()
	defer func() {
		if err != nil {
			// no need to check for drop - the state is already
			// dropped - we just need to decref here.
			// XXX locking?
			pyobj.deactivate()
		}
	}()
	if !doload {
		loading := pyobj.loading
		pyobj.mu.Unlock()

		// someone else is already activated/activating the object.
		// wait for its loading to complete and we are done.
		select {
		case <-ctx.Done():
			return ctx.Err()	// XXX err ctx
		case <-loading.ready:
			return loading.err	// XXX err ctx?
		}
	}

	// we become responsible for loading the object
	loading := &loading{ready: make(chan struct{})}
	pyobj.loading = loading	// XXX assert before it was = nil ?
	pyobj.mu.Unlock()

	pyclass, pystate, serial, err := pyobj.jar.loadpy(ctx, pyobj.oid)
	if err == nil && pyclass != pyobj.pyclass {
		// complain pyclass changed
		// (both ref and object data uses pyclass so it indeed can be different)
		err = &wrongClassError{want: pyobj.pyclass, have: pyclass} // XXX + err ctx
		pystate = nil
	}

	pyobj.mu.Lock()

	pyobj.serial  = serial
//	pyobj.pystate = pystate

	if err == nil {
		err = pyobj.instance.PySetState(pystate)	// XXX err ctx
	}
	if err != nil {
		pyobj.instance.DropState()
	}

	// XXX unlock

	loading.err = err
	close(loading.ready)
	return err	// XXX err ctx
}


// PDeactivate implements Object.
func (pyobj *pyObject) PDeactivate() {
	drop := pyobj.pdeactivate()
	if !drop {
		return
	}

	// we have to drop pyobject state
	// XXX locking?
//	pyobj.pystate = nil
	pyobj.instance.DropState()
	pyobj.loaderr = nil
	pyobj.ready   = make(chan struct{})
}

// XXX pyobj.PInvalidate()	= deactivate without checking if state != modified
func (pyobj *pyObject) PInvalidate() {
	// XXX panic if refcnt != 0  (object being used)
	pyobj.instance.DropState()
	pyobj.loaderr = nil
	pyobj.ready   = make(chan struct{})
}
